import 'clean-stack'
import { resolve, dirname, relative, join } from 'path'
import { EventEmitter } from 'fbemitter'
import lodash from 'lodash'
import findUp from 'find-up'
import vm from 'isomorphic-vm'
import { promiseHash, mixinPropertyUtils, lazy, query, strings } from 'skypager-util'
import Collection, { Document, mount as mountCollection } from 'skypager-documents'
import config from './config/index'
import middleware from './middleware'
import createDocumentTree from './tree'
import Cache from './cache'

const { sortBy, omit, isError, attempt, isFunction, defaults, result, mapValues, isArray, get, chain } = lodash
const { singularize } = strings

const { keys, values, assign } = Object

export const DefaultOptions = {
  cwd: process.cwd(),
  target: 'node',
  paths: {
    source: 'src',
    output: 'dist',
    bundles: 'dist/bundles',
    ast: 'dist/ast',
    docs: 'docs',
    data: 'data',
    logs: 'log',
    temp: 'tmp',
    main: 'skypager.js',
    manifest: 'package.json',
  }
}

export const chainSources = [
  'scopes.available',
  'constructor.chainProps',
  'options.chainProps',
]

export const querySources = [
  'scopes.available',
  'constructor.queryProps',
  'options.queryProps',
]

export class Project {
  /**
 * Create a Skypager Project wrapper around a package folder.
 *
 * @param cwd {String} an absolute path for the Project. Should be the folder which has the package.json in it.
 * @param options {Object} the options Object
 * @param options.filesystem {Boolean} use the portfolios file system adapters
 * @type {Object}
 */
  constructor (cwd = process.cwd(), options = {}, context = {}) {
    mixinPropertyUtils(this)
    attachEmitter(this)
    const project = this

    this.cwd = cwd

    this.instanceId = result(options, 'id', () => `${this.namespace}-${ Math.floor(Date.now() / 1000)}`)

    this.hide('current', {
      collection: 'project',
      middleware: 'project',
      configuration: 'project',
    })

    this.hideGetter('framework', () => context.framework || context.portfolio)
    this.getter('portfolio', () => context.portfolio)

    this.hide('manifest', options.manifest || this.readJsonSync(
      this.join(get(options, 'paths.manifest', 'package.json'))
    ))

    const computedOptions = defaults(options,
      this.get('manifest.skypager'),
      DefaultOptions
    )

    this.hide('options', omit(computedOptions, 'manifest', 'portfolio', 'fs', 'fsx'))

    this.hide('middleware', middleware(this))

    require('./logger').attach(this, this.getOption('logging', {
      console: true
    }))

    this.hide('configuration', options.configuration || {})

    this.hide('cache', new Cache())

    require('./middlewares/registries').attach(this, {
      collections: {
        type: 'Directory',
        wrapper: function(cfg, id) {
          return project.cache.fetch(`collections:${id}:${project.instanceId}:${cfg.cwd}`, () => project.mountCollection(cfg.cwd, cfg))
        },
      },
      middlewares: {
        type: 'Directory'
      },
      selectors: {
        type: 'Directory',
        wrapper: function(fn, id) {
          return (stayChained) => project.cache.fetch(`selectors:${id}:${stayChained}:${project.cacheKey}`, () => (
            stayChained ? fn(project.chain) : fn(project.chain).value()
          ))
        }
      },
      configurations: {
        type: 'Directory',
        wrapper: function(cfg, id) {
          return project.cache.fetch(`configurations:${id}:${project.instanceId}`, () => {
            return cfg.call(project, project.configure({scope: id})).getConfig()
          })
        },
      },
      helpers: {
        type: 'Context',
        context: require.context('./helpers', false, /\.js$/),
        fallback(requestId) {
          return project.portfolio.helpers.lookup(requestId)
        }
      }
    })

    this.register.collection('project', {
      cwd: this.cwd,
      sync: !!options.sync,
    })

    this.register.collection('source', {
      cwd: this.paths.source,
      sync: !!options.sync,
      exclude: ['**/*.spec.*'],
    })

    this.register.collection('asts', {
      cwd: this.paths.ast,
      sync: false,
      exclude: ['**/*.spec.*'],
    })

    this.register.collection('logs', {
      cwd: this.paths.logs,
      sync: false,
    })

    this.register.collection('bundles', {
      cwd: this.paths.bundles,
      sync: false,
      type: 'script',
    })

    this.register.collection('output', {
      cwd: this.paths.output,
      sync: false,
      exclude: ['**/ast/**'],
    })

    if (options.sync) {
      const cachedCollection = this.collection.cached
      this.emit('collectionDidLoad', cachedCollection)
    }

    this.lazy('files', () => this.collection.files.map((f) => this.loadFile(f)))
    this.lazy('documents', () => this.collection.withDocuments.documents.map(d => this.loadDocument(d)))
    this.lazy('exportableDocumentsTree', () => this.createExportableDocumentsTree())
    this.lazy('cacheKey', () => this.calculateCacheKey())

    this.get('constructor.collectionDelegators', Project.collectionDelegators).forEach(meth => (
      this.hideGetter(meth, () => project.result(`collection.${meth}`))
    ))
  }

  get lookup() {
    const p = this

    return p.registries.available
      .reduce((memo, prop) => ({
        ...memo,
        [singularize(prop)](...args) {
          return p.registries.lookup(prop).lookup(...args)
        }
      }), {})
  }

  get register() {
    const p = this

    return p.registries.available
      .reduce((memo, prop) => ({
        ...memo,
        [singularize(prop)](...args) {
          return p.registries.lookup(prop).register(...args)
        }
      }), {})
  }

  document (documentId) {
    const doc = this.get(['documentsMappedById', documentId])
    return doc ? doc.ready : doc
  }

  get documentsTree() {
    return createDocumentTree(this, {
      ready: false,
    })
  }

  get loadedDocumentsTree() {
    return createDocumentTree(this, {
      ready: true,
    })
  }

  createExportableDocumentsTree() {
    return mapValues(this.loadedDocumentsTree, (v) => (
      v.createExportable ? v.exportables : v
    ))
  }

  configure (options = {}) {
    if (typeof options === 'string') {
      options = { scope: `${options}` }
    }

    const {
      scope = this.current.configuration || 'project',
      api = require('./config').api,
    } = options

    const current = this.configuration[scope]

    if (current) {
      return api({
        scope: this,
        ...options,
        history: current.history,
      })
    }

    return this.configuration[scope] = api({
      features: {},
      reducers: {},
      scope: this,
      ...options,
      history: [],
    })
  }

  get context() {
    return this.cwd
  }

  get root() {
    return this.cwd
  }

  get parentFolder() {
    return dirname(this.cwd)
  }

  get id() {
    return `${this.instanceId}-${this.get('manifest.version')}`
  }

  get name() {
    return this.get('options.name') || this.get('manifest.name') || this.constructor.name.toLowerCase()
  }

  get paths() {
    const project = this

    return {
      cwd: this.cwd,
      ...this.chain
        .get('options.paths', {})
        .defaults(DefaultOptions.paths)
        .mapValues((p) => (project.resolve(p)))
        .value()
    }
  }

  /**
   * A namespace can be set in the Project options, it will default
   * to the package.json scope if it exists (e.g @skypager/package) otherwise
   * the name of the Projects parent folder will be used.
   */
  get namespace() {
    const name = this.name

    return this.result('options.namespace', () => (
      name.match(/^\@\w+\//)
        ? name.split('/')[0].replace(/\^\@/,'')
        : name.split('-')[0]
      )
    )
  }

  get packageMain() {
    return this.get('manifest.main', 'index.js')
  }

  get skypagerMain() {
    return this.get('manifest.skypager.main', 'skypager.js')
  }

  get isCli() {
    return !!this.get('environment.SKYPAGER_CLI', false)
  }

  get isServer() {
    return !!this.get('environment.SKYPAGER_SERVER', false)
  }

  get isElectron() {
    return !!this.get('environment.SKYPAGER_ELECTRON', false)
  }

  get chain() {
    return chain(this)
  }

  get collection() {
    return this.collections.lookup(this.current.collection)
  }



  get availableHelpers() {
    return this.chain.at('helpers.available','portfolio.helpers.available')
      .flatten()
      .uniq()
      .sort()
      .value()
  }

  get helpersMap() {
    const helpersRoot = this.getOption('helpersRoot', 'src')
    const helpers = this.getOption('helpers', [])

    const normalized = helpers.map(entry => (
      typeof entry === 'string'
        ? {name: entry, pattern: `${helpersRoot}/${entry}/:name*`}
        : entry
    ))

    return normalized.reduce((memo, cfg) => ({
      ...memo,
      [cfg.name]: {...cfg, documents: this.routeDocuments(cfg.pattern)}
    }), {})
  }

  route(pattern, chain) {
    return this.routeDocuments(pattern, chain)
  }

  routeDocuments(pattern, chain) {
    const project = this
    const collection = this.collection

    const results = collection.routers
      .documents(pattern)
      .map(result => {
        const id = result.path[0]
        const doc = result.path[1]
        const file = doc.file
        const idParts = id.split('/')

        return {
          id,
          doc,
          idParts,
          baseFolder: idParts[0],
          categoryFolder: idParts[1] || idParts[0],
          meta: get(result, 'result', {}),
          file: {
            path: file.path,
            ...project.wrapFile(file),
          }
        }
      })

    return chain ? lodash.chain(results) : results
  }

  routeFiles(pattern, chain) {
    const project = this
    const collection = this.collection

    const results = collection.routers.files(pattern).map(result => {
      const file = result.path[1]
      const id = file.relative
      const idParts = id.split('/')

      return {
        id,
        file,
        idParts,
        baseFolder: idParts[0],
        categoryFolder: idParts[1] || idParts[0],
        meta: get(result, 'result', {}),
        ...(project.wrapFile(file))
      }
    })

    return chain ? lodash.chain(results) : results
  }

  get vm() {
    return vm
  }

  get sandbox () {
    return this.createSandbox()
  }

  get module() {
    return require.cache[__non_webpack_require__.resolve(this.paths.manifest)]
  }

  resolvePromiseHash(object = {}) {
    return promiseHash(object)
  }

  createSandbox (locals = {}) {
    return {
      project: this,
      Skypager: this.portfolio,
      lodash,
      fs: this.fsx,
      ...locals,
      console
    }
  }

  loadFile (file) {
    return file
  }

  loadDocument (doc) {
    return doc.decorated
  }

  wrapFile (file) {
    const project = this
    const { path } = file

    return {
      contents() {
        return project.readFileSync(path).toString()
      },

      createScript(options = {}) {
        // find a way to do this
        const code = options.compile
          ? file.contents.toString()
          : file.contents.toString()

        return new proj.vm.Script(code, {
          filename: path,
          ...options,
        })
      },

      getFile() {
        return file
      },

      json() {
        return project.readJsonSync(path)
      },

      read() {
        return project.readFileAsync(path).then(file => file.toString())
      },

      readJson() {
        return project.readJsonAsync(path)
      },
    }
  }
  /**
   * Get information about this project's parent. It will look up the path for the closest package.json
   * @return {Object} An object which provides access to the package.json path, folder path, and contents
   */
  get parentModule() {
    const proj = this

    return {
      get path() {
        return proj.findUp('package.json', {
          cwd: proj.parentFolder,
        })
      },

      get folder() {
        return dirname(proj.parentModule.path)
      },

      get manifest() {
        return proj.framework.readJsonSync(
          proj.parentModule.path
        )
      },

      get projectLoader() {
        return (options = {}) => proj.portfolio.load(proj.parentModule.folder, options)
      },

      get project() {
        return proj.parentModule.projectLoader({
          sync: false
        })
      }
    }
  }

  get argv() {
    return {
      ...require('minimist')(process.argv),
      ...this.getOption('argv', {})
    }
  }

  get env() {
    const env = this.environment
    return env.SKYPAGER_ENV ? env.SKYPAGER_ENV : env.NODE_ENV
  }

  get environment () {
    return {
      NODE_ENV: 'development',
      ...this.getOption('env', process.env),
    }
  }

  calculateCacheKey() {
    const wrappers = this.chain.get('collection.importer.fileWrappersSync')

    this.maxUpdatedAt = wrappers.map(f => (
      f.stat && f.stat.mtime && f.stat.mtime.getTime()
    )).max().value()

    this.filesCount = wrappers.size().value()

    return `${this.id}:${this.maxUpdatedAt}:${this.filesCount}`
  }

  get chains() {
    const project = this

    const chains = this.chain
      .at(...chainSources)
      .flatten()
      .compact()
      .uniq()
      .reduce((memo, prop) => {
        lazy(memo, prop, ()=> chain(project.get(prop)))
        return memo
      },{})

    return chains.value()
  }

  get queries() {
    const project = this

    const queries = this.chain
      .at(...querySources)
      .flatten()
      .compact()
      .uniq()
      .reduce((memo, prop) => {
        memo[prop] = (...args) => query(project.get(prop), ...args)
        return memo
      },{})

    return queries.value()
  }

  get gitignores() {
    return this.getOption('ignorePaths', ['tmp/', 'log/', 'dist/', '.DS_Store'])
      .concat(this.gitignoreContents)
      .map(p => `!${p}`)
  }

  get gitignoreContents() {
    return this.attemptReadSync('.gitignore', '').split("\n")
    .filter(p => p && p.length > 2 && !p.trim().startsWith('#'))
  }

  get skypagerIgnoreContents() {
    return this.attemptReadSync('.skypagerignore', '').split("\n")
    .filter(p => p && p.length > 2 && !p.trim().startsWith('#'))
  }

  /**
    PATH HELPERS

    Helpers to make working with project paths easier
  */
  join(...args) { return join(this.cwd, ...args) }

  relative(...args) { return relative(this.cwd, ...args) }

  resolve(...args) {
    return resolve(this.cwd, ...args)
  }

  /**
   * Access one of the project options, providing a default value should one
   * not already be set.

   * @param          {String} value        an object path including something like whatever.nested.value
   * @param          {Any} defaultValue    an optional default value should the requested value not be defined
   */
  getOption(key, defaultValue) {
    return this.get(`options.${key}`, this.getConfig(key, defaultValue))
  }

  getConfig(key, defaultValue) {
    return this.get(`configuration.${this.current.config}.${key}`, defaultValue)
  }
  /**
   * Access one of the project options, invoking a function if found. Providing a default value should one
   * not already be set, including a function which will be lazily evaluated.

   * @param          {String} value        an object path including something like whatever.nested.value
   * @param          {Any} defaultValue    an optional default value, or function returning one,
   *                                       should the requested value not be defined
   */
  resultOption(value, defaultValue) {
    return this.result(`options.${value}`, defaultValue)
  }

  attemptReadSync(path, fallbackValue = '') {
    const result = attempt(
      () => this.fs.readFileSync(this.join(path)).toString()
    )

    return isError(result)
      ? fallbackValue
      : result
  }

  use (middleware) {
    if (isFunction(middleware)) {
      this.middleware(middleware)
    }

    return this
  }

  start (options = {}, callback) {
    if (isFunction(callback)) {
      return this.middleware.run(options, this.sandbox, callback)
    } else {
      return this.startAsync(options, this.sandbox)
    }
  }

  startAsync (options = {}) {
    return new Promise((resolve,reject) => {
      this.middleware.run(options, this.sandbox, (err, ...callbackArgs) => {
        if (err) { reject(err) } else { resolve(callbackArgs) }
      })
    }).catch(error => {
      console.log('Error in start async method', this, error)
    })
  }

  partialMixin(InterfaceDefinition, ...args) {
    const project = this

    chain(InterfaceDefinition)
      .pickBy(v => typeof v === 'function')
      .mapValues((fn) => fn.bind(project, ...args))
      .mapValues((fn, methodName) => project.hide(methodName, fn, true))

    return this
  }

  mixin (InterfaceDefinition = {}) {
    assign(this, mapValues(InterfaceDefinition, (fn) => fn.bind(this)))
    return this
  }

  mountCollection (cwd = this.cwd, options = {}) {
    const project = this

    options = {
      importerOptions: project.get('options.importer', project.get('options.importerOptions', {})),
      project: project,
      documentClass: Document,
      ...options,
      exclude: [
        ...options.exclude || [],
        ...options.ignore || [],
        ...project.gitignores || [],
        ...project.skypagerIgnoreContents || [],
      ],
    }

    return mountCollection(cwd, {
      ...options,
      collection: this,
    })
  }

  findUp(fileName, options = {}) {
    return findUp.sync(fileName, {
      cwd: this.parentFolder,
      ...options,
    })
  }

  select(selector, chain = false) {
    if (isFunction(selector)) {
      return selector.call(this, this.chain)(chain)
    } else {
      if (this.selectors.available.indexOf(selector) !== -1) {
        return this.selectors.lookup(selector)(chain)
      } else {
        const result = this.selectChain(selector)
        return chain ? result : result.value()
      }
    }
  }

  selectChain(...paths) {
    const set = (obj,path,val) => {
      lodash.set(obj,path,val)
      return obj
    }

    let index = 0
    paths = sortBy(paths, (p) => p.length)

    return lodash.chain(this.at(paths)
      .reduce((result,value) => {
        const key = paths[index++]
        return set(result, key, value)
      }, {}))
  }

  selectByObject (pathObject ) {
    const keys = keys(pathObject)
    const paths = values(pathObject)

    return paths.map(p => this.select(p))
      .reduce(
        ((memo,val,index) => set(memo, keys[index], val)), {}
      ).value()
  }

  setupScopes (config = {}) {
    if (isArray(config)) {
      this.scope(config, new RegExp(`${config}\/.*\.\w+$`))
    } else {
      keys(config).forEach(scopeName => {
        const scopeDef = config[scopeName]

        if (typeof scopeDef === 'string') {
          this.scope(scopeName, new RegExp(`${scopeDef}\/.*\.\w+$`))
        }

        if (typeof scopeDef === 'object') {
          this.scope(scopeName, scopeDef.pattern, scopeDef)
        }
      })
    }
  }

  /**
   * Create a lodash chain by picking some properties from the project.
   * @param  {Array[String]} properties List of properties you want to be a part of the chain
   * @return {Chain}           The lodash chain whose initial value consists of the properties given
   */
  pick (properties) {
    return this.chain.pick(properties)
  }

  /**
   * Create a lodash chain with the given object paths.
   */
  at (...paths) {
    return this.chain.at(...paths)
  }

  resolveRequest(request, options = {}) {
    if (!options.async) {
      return Promise.resolve(this.resolveSync(request, options)).catch(error => {
        console.log('ERROR in Resolve Request', error)
      })
    } else {
      return new Promise((resolve, reject) => {
        resolve(this.resolveSync(request, options))
      }).catch(error => {
        console.log('ERROR in Resolve Request', error)
      })
    }
  }

  resolveRequestSync (request, options = {}) {
    if (typeof request === 'object') {
      return request
    } else if (typeof request === 'string') {
      return {
        request
      }
    }
  }

  static collectionDelegators = [
    'childFolderNames',
    'childFolderPaths',
    'subFolderNames',
    'subFolderPaths',
    'rootFiles',
    'groupedByChildFolder',
    'groupedByChildFolderAndExtension',
    'relativeFilePaths',
    'relativeSubFolderPaths',
    'getFileByRelativePath',
    'filesIndex',
    'relativeObjectPaths',
    'subFolderObjectPaths',
    'routers',
    'documentsMap',
    'documentsMappedById',
    'documentEntryPoints',
    'depths',
    'gitInfo',
  ]

  static chainProps = [
    'files',
    'documents',
    'childFolderPaths',
    'childFolderNames',
    'subFolderNames',
    'subFolderPaths',
  ]

  static queryProps = [
    'files',
    'documents',
    'childFolderPaths',
    'childFolderNames',
    'subFolderNames',
    'subFolderPaths',
  ]

  static get Collection() {
    return Collection
  }

  static get Document() {
    return Document
  }

  static load (cwd = process.cwd(), options = {}, context = {}) {
    return new this(cwd, options, context)
  }

  static loadSync(cwd = process.cwd(), options = {}, context = {}) {
    return this.load(cwd, {
      ...options,
      sync: true
    }, context)
  }
}

export default Project

export function load(cwd = process.cwd(), options = {}, context = {}) {
  return new Project(cwd, options, context)
}

export function attachEmitter(project) {
    project.hide('emitter', new EventEmitter())
    project.hide('on', project.emitter.addListener.bind(project.emitter))
    project.hide('once', project.emitter.once.bind(project.emitter))
    project.hide('off', project.emitter.removeCurrentListener.bind(project.emitter))
    project.hide('emit', project.emitter.emit.bind(project.emitter))
    project.hide('trigger', project.emitter.emit.bind(project.emitter))
}

export function attachRegistry (project, name, options = {}) {
  if (project.registries.available.indexOf(name) >= 0) {
    return project.registries.lookup(name)
  }

  project.registries.register(name, () => create({
    name,
    type: 'Directory',
    ...options,
  }))

  if (options.prop !== false && typeof project[name] === 'undefined') {
    project.getter(name, () => project.registries.lookup(name))
  }

  if (options.prop !== false && typeof project[singularize(name)] === 'undefined') {
    project.getter(singularize(name), (...args) => project[name].lookup(...args))
  }

  return project
}
